/****************************************************************************
 * NCSA HDF                                                                 *
 * Software Development Group                                               *
 * National Center for Supercomputing Applications                          *
 * University of Illinois at Urbana-Champaign                               *
 * 605 E. Springfield, Champaign IL 61820                                   *
 *                                                                          *
 * For conditions of distribution and use, see the accompanying             *
 * hdf/COPYING file.                                                        *
 *                                                                          *
 ****************************************************************************/

#ifdef RCSID
static char RcsId[] = "@(#)$Revision: 1.134 $";
#endif

/* $Id: mfsd.c,v 1.134 2001/09/26 20:50:49 bmribler Exp $ */
#ifdef HAVE_PABLO
#define PABLO_mask ID_mfsd_c
#include "ProcIDs.h"
#endif

/******************************************************************************
file - mfsd.c

  This file contains the HDF/netCDF based multi-file interface for SDSs

  All of the routines that make up this interface have names beginning
  with SD.  Routines beginning with SDI are internal routines and
  should not be used outside of this module.

  Defining SDDEBUG will print status messages to stderr

Reading interface:
------------------

fid    = SDstart(file name, access);

        --- get the number of data-sets in the file ---
num    = SDnumber(fid);
        
sdsid  = SDselect(fid, i, ...);   0 <= i < num

        --- return the name, rank, dimsizes, #of attr, datatype ---
status = SDgetinfo(sdsid, ...); 

status = SDreaddata(sdsid, ...);

status = SDgetrange(sdsid, ...);

status = SDend(fid);

status = SDisdimval_bwcomp(dimid);

status = SDcheckempty(sdsid, emptySDS);

 NOTE: This file needs to have the comments cleaned up for most of the
       functions here. -GV 9/10/97

******************************************************************************/

#include "local_nc.h"
#ifdef VMS
#include <stat.h>
#endif

#ifdef HDF
#include "mfhdf.h"
#include "hfile.h"
#include "hchunks.h" /* include here for now */

/* for Chunk debugging */
/*
#define CHK_DEBUG
*/

/* Local function prototypes */

PRIVATE NC * SDIhandle_from_id 
    (int32 id, intn typ);

PRIVATE NC_var *SDIget_var
    (NC *handle, int32 sdsid);

#ifdef NOTDEF
PRIVATE NC_dim * SDIget_dim
    (NC *handle, int32 id);

PRIVATE intn SDIputattr 
    (NC_array **ap, const char *name, int32 nt, intn count, const void * data);

PRIVATE int32 SDIgetcoordvar 
    (NC *handle, NC_dim *dim, int32 id, int32 nt);

PRIVATE int32 SDIfreevarAID 
    (NC * handle, int32 index);

PRIVATE intn SDIapfromid
    (int32 id, NC ** handlep, NC_array *** app);

#endif

/* Whether we've installed the library termination function yet for this interface */
PRIVATE intn library_terminate = FALSE;

/******************************************************************************
 NAME
	SDgetcompress -- Retrieves compression information of a dataset

 DESCRIPTION
    This routine uses HCgetcompress to retrieve the compression type
    and the compression information of the identified dataset.

 RETURNS
    SUCCEED/FAIL

 MODIFICATION
    July 2001: Added to fix bug #307 - BMR 

******************************************************************************/ 
intn 
MySDgetcompress(int32     id,     /* IN: dataset ID */
              comp_coder_t*    comp_type,   /* OUT: the type of compression */
              comp_info* c_info,  /* OUT: ptr to compression information 
				structure for storing the retrieved info */
              int32*      length) /* OUT: compressed length */
{
    CONSTR(FUNC, "SDgetcompress");    /* for HGOTO_ERROR */
    NC        *handle;
    NC_var    *var;
    intn       status = FAIL;
    intn       ret_value = SUCCEED;

#ifdef SDDEBUG
    fprintf(stderr, "SDgetcompress: I've been called\n");
#endif /* SDDEBUG */

#ifdef HAVE_PABLO
    TRACE_ON(PABLO_mask, ID_SDgetcompress );
#endif

    /* clear error stack */
    HEclear();

    if(comp_type == NULL || c_info == NULL)
	HGOTO_ERROR(DFE_ARGS, FAIL);

    handle = SDIhandle_from_id(id, SDSTYPE);
    if(handle == NULL || handle->file_type != HDF_FILE)
	HGOTO_ERROR(DFE_ARGS, FAIL);
    if(handle->vars == NULL)
	HGOTO_ERROR(DFE_ARGS, FAIL);

    var = SDIget_var(handle, id);
    if(var == NULL) HGOTO_ERROR(DFE_ARGS, FAIL);

#ifdef SDDEBUG
    printf("SDgetcompress(): var->data_ref=%d, var->aid=%d\n",(int)var->data_ref, (int)var->aid);
#endif /* SDDEBUG */
    if(!var->data_ref) 
	HGOTO_ERROR(DFE_ARGS, FAIL);

    /* use lower-level routine to get the compression information */
    status = MyHCgetcompress(handle->hdf_file, var->data_tag, var->data_ref, 
		comp_type, c_info, length);
    if(status==FAIL) HGOTO_ERROR(DFE_INTERNAL, FAIL);

done:
    if (ret_value == FAIL)
      { /* Failure cleanup */

      }
    /* Normal cleanup */

#ifdef HAVE_PABLO
    TRACE_OFF (PABLO_mask, ID_SDgetcompress);
#endif
    return ret_value;    
} /* SDgetcompress */
/******************************************************************************
 NAME
	SDIhandle_from_id -- get the handle from this object

 DESCRIPTION
    Map an ID to the handle for this file

 RETURNS
    NULL or the handle

******************************************************************************/
PRIVATE NC *
SDIhandle_from_id(int32 id, /* IN: an object (file, dim, dataset) ID */
                  intn  typ /* IN: IN: the type of ID this is */)
{
    int32 tmp;
    NC   *ret_value = NULL;

    /* check that it is the proper type of id */
    tmp = (id >> 16) & 0xff;
    if(tmp != typ)
      {
        ret_value = NULL;
        goto done;
      }

    /* get the file from top 8 bits*/
    tmp = (id >> 24) & 0xff;
    ret_value = NC_check_id((int)tmp);

done:
    if (ret_value == NULL)
      { /* Failure cleanup */

      }
    /* Normal cleanup */

    return ret_value;
} /* SDIhandle_from_id */


/******************************************************************************
 NAME
	SDIget_var -- get the variable record

 DESCRIPTION
    Map an ID and a handle to the NC_var object for this dataset

 RETURNS
    NULL or the variable object

******************************************************************************/
PRIVATE NC_var *
SDIget_var(NC   *handle, /* IN: the handle for this file */
           int32 sdsid   /* IN: a dataset ID */)
{
    int32      varid;
    NC_array **ap = NULL;
    NC_var    *ret_value = NULL;

    /* varid is low 16bits of sdsid */
    varid = sdsid & 0xffff;

    if(handle->vars != NULL 
       && varid >= 0 
       && varid < handle->vars->count) 
      {
        ap = (NC_array **)handle->vars->values;
        ap += varid;
      } 
    else 
      {
        ret_value = NULL;
        goto done;
      }
    
    ret_value = ((NC_var *)*ap);

done:
    if (ret_value == NULL)
      { /* Failure cleanup */

      }
    /* Normal cleanup */

    return ret_value;    
} /* SDIget_var */

#endif /* HDF */
