/*
=head1 NAME

hdfzip - apply internal HDF compression to an HDF file

=head1 SYNOPSIS

hdfzip -i infile -o outfile [-v] [-c comp_type comp_info] [-f comp_file]

=head1 DESCRIPTION

hdfzip uses HDF internal compression methods to generate a compressed HDF file.

=head1 ARGUMENTS

=over 4

=item -i infile

Input HDF file.

=item -o outfile

Output HDF file

=item -v

Verbose mode. Writes various information messages to stderr.

=item -c comp_type comp_info

Numerical code indicating compression type:
  1 = RLE
  3 = Skipping Huffman
  4 = Gzip (LZW)

Comp_info is additional information.  It is not needed for RLE.

For Skipping Huffman, it is the skip factor.

For gzip it is the compression factor, 1 to 9, with 9 being
maximum compression.  Unless space is the only consideration, a
factor of 1 is recommended, as it goes much faster with only a
few percent increase in size.

This applies the same compression technique to all SDSs in the
file.  

=item -f comp_file

Alternatively, a compression control file can be specified.
See the FILES section below for format

=back

=head1 FILES

The compression info file has information for individual SDS's,
one on each line.  The format is:
  number_of_lines
  sds_name comp_type comp_info rank chunk chunk ...
  sds_name comp_type comp_info rank chunk chunk ...
  ...
where the number of chunk parameters is the rank of the SDS.

Here is an example for MOD021KM:

  2
  EV_1KM_RefSB 4 1 3 1 203 677
  EV_1KM_RefSB_Uncert_Indexes 4 1 3 1 203 677

This applies gzip with minimal (fastest) compression to EV_1KM_RefSB
and EV_1KM_RefSB_Uncert_Indexes, where each chunk is a 203x677 pixel
tile in each band.

=head1 BUGS

Does not compute proper compressed size for chunked datasets.
(Ha! Like that's the only bug...)

=cut
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "hdf.h"
#include "mfhdf.h"
#include "hfile.h"
#include "hcompi.h"     /* Internal definitions for compression */

#define MAX_RANK 1024

typedef struct {
    char *sds_name;
    int  comp_type;
    int  comp_info;
    int  chunk[MAX_RANK];
} sds_comp_t;

void usage();
int32 *transfer_sds(char *, char *, sds_comp_t *, int, int, int, int, int);
sds_comp_t *read_compression_info(char *filename);

main(int argc, char **argv)
{
    char *infile = NULL, *outfile = NULL;
    int  i, j, rc, verbose = 0;
    int  compression = 0;
    int  check_fill = 0;
    int  compression_info = 1;
    int  chunk = 0;
    char *buf, *p_start, *p_end;
    sds_comp_t *sds_comp;

    sds_comp = (sds_comp_t *)malloc(sizeof(sds_comp_t) * argc);
    for (i = 0; i < argc; i++){
        sds_comp[i].sds_name = NULL;
        sds_comp[i].comp_type = 0;
        for (j = 0; j < MAX_RANK; j++) {
            sds_comp[i].chunk[j] = 0;
        }
    }

    for (i = 1, j = 0; i < argc; i++) {
        if (strcmp(argv[i], "-i") == 0) {
            infile = argv[++i];
        }
        else if (strcmp(argv[i], "-o") == 0) {       
            outfile = argv[++i]; 
        }
        else if (strcmp(argv[i], "-v") == 0) {       
            verbose = 1;
        }
        else if (strcmp(argv[i], "-F") == 0) {       
            check_fill = 1;
        }
        else if (argv[i][0] == '-' && isdigit(argv[i][1])) {       
            compression = atoi(argv[i]+1); 
            if (i < (argc-1) && isdigit(argv[i+1][0])) {
                i++;
                compression_info = atoi(argv[i]); 
            }
        }
        else if (strcmp(argv[i], "-C") == 0) {       
            chunk = 1;
        }
        else if (strcmp(argv[i], "-f") == 0) {       
            sds_comp = read_compression_info(argv[++i]);
            compression = 1;
        }
        else if (argv[i][0] == '-') {
            usage();
        }
        else {
            sds_comp[j].sds_name = argv[i];
            sds_comp[j].comp_type = compression;
            sds_comp[j].comp_info = compression_info;
            j++;
        }
    }
    if (! compression) {
        free(sds_comp);
        sds_comp = NULL;
    }
    else if (sds_comp[0].sds_name == NULL) {
        sds_comp[0].comp_type = compression;
        sds_comp[0].comp_info = compression_info;
    }
    if (infile == NULL || outfile == NULL) {
        usage();
    }
    /* transfer_vgroups actually transfers everything:  SDS, vgroups, vdata */
    rc = transfer_vgroups(infile, outfile, sds_comp, check_fill, chunk, verbose);
    exit(0);
}
/* 
 * Compute_chunk_lengths:  attempt to compute reasonable chunking for
 * compression.  Unfortunately, an abysmal failure.  Better to use the
 * -f option to read in a file with compression and chunking info.
 */
int 
compute_chunk_lengths(int rank, int *dim_sizes, int *chunk_lengths) {
    int max_dim_size = 0;
    int i, split;
    if (rank <= 1) {
        return 0;
    }
    for (i = 0; i < rank; i++) {
        max_dim_size = (dim_sizes[i] > max_dim_size) ? dim_sizes[i] : max_dim_size;
    }
    for (i = 0; i < (rank-1); i++) {
        if (dim_sizes[i] < 25 && ((double)dim_sizes[i]/(double)max_dim_size) < 0.01) {
            chunk_lengths[i] = 1;
        }
        else {
            split = (int)pow(2.,(int)log10(dim_sizes[0])-2);
            if (split <= 0) split = 1;
            chunk_lengths[i] = dim_sizes[i] / split;
        }
    }
    split = (int)pow(2.,(int)log10(dim_sizes[rank-1])-1);
    if (split <= 0) split = 1;
    chunk_lengths[rank-1] = dim_sizes[rank-1] / split;
    return 1;
}
int 
copy_sds_attributes(int32 id_in, int32 id_out, int n_attrs, int verbose)
{
    int i, n_values;
    int32 data_type;
    char attr_name[1024];
    int  attr_size;
    char *values;
    /* Loop through attributes in input SDS */
    for (i = 0; i < n_attrs; i++) {
        if (SDattrinfo (id_in, i, attr_name, &data_type, &n_values) == FAIL) {
            fprintf(stderr, "Cannot get info for attribute number %d\n", i);
            exit(30);
        }
        /* Allocate space for attributes */
        attr_size = data_type_size(data_type);
        if ((values = (char *)malloc(attr_size * n_values)) == NULL) {
            fprintf(stderr, "Error allocating %d values of size %d for attribute %s",
                n_values, attr_size, attr_name);
            exit(31);
        }
        /* Read attributes from input SDS */
        if (SDreadattr(id_in, i, values) == FAIL) {
            fprintf(stderr, "Cannot read attribute %s\n", attr_name);
            exit(32);
        }

        /* Put attributes into output SDS */
        if (verbose) fprintf(stderr, "Copying attribute %s...\n", attr_name);
        if (SDsetattr(id_out, attr_name, data_type, n_values, values) == FAIL) {
            fprintf(stderr, "Cannot write attribute %s\n", attr_name);
            exit(33);
        }
    }
    return 1;
}
int
copy_vdata_attribute(int32 in, int32 out, int field, int attr, int verbose)
{
    char attr_name[1024];
    int32 n_values, attr_size, attr_type;
    char *values;
    /* Get attribute information */
    VSattrinfo(in, field, 0, attr_name, &attr_type, &n_values, &attr_size);
    /* Allocate space for attribute values */
    if ((values = (char *)malloc(attr_size * n_values)) == NULL) {
        fprintf(stderr, "Cannot allocate %d values of size %d for attribute %s",
            n_values, attr_size, attr_name);
        exit(31);
    }
    /* Read attribute from input object */
    if (VSgetattr(in, field, 0, values) == FAIL) {
        fprintf(stderr, "Cannot read attribute %s\n", attr_name);
        exit(32);
    }
    /* Write attribute to output object */
    if (VSsetattr(out, field, attr_name, attr_type, n_values, values) == FAIL) {
        fprintf(stderr, "Cannot write attribute %s\n", attr_name);
        exit(32);
    }
    return 1;
}
    
int 
data_type_size(int32 data_type) {
    switch (data_type) {
        case DFNT_CHAR8:
        case DFNT_UCHAR8:
        case DFNT_INT8:
        case DFNT_UINT8:
            return 1;
        case DFNT_INT16:
        case DFNT_UINT16:
            return 2;
        case DFNT_INT32:
        case DFNT_UINT32:
        case DFNT_FLOAT32:
            return 4;
        case DFNT_FLOAT64:
            return 8;
        default:
            fprintf(stderr, "Unknown datatype %d\n", data_type);
            exit(20);
    }
    return 0;
}
/* Retrieve desired compression/chunking information for specified SDS */
int
get_compression_info(char *sds_name, sds_comp_t *sds_comp, int *compression_info, int *chunk)
{
    int i, j;
    if (sds_comp == (sds_comp_t *)NULL) return 0;
    for (i = 0; sds_comp[i].sds_name != NULL; i++) {
        if (strcmp(sds_name, sds_comp[i].sds_name) == 0) break;
    }
    if (i == 0 || sds_comp[i].sds_name != NULL) {
        for (j = 0; j < MAX_RANK && sds_comp[i].chunk[j]; j++) {
            chunk[j] = sds_comp[i].chunk[j];
        }
        *compression_info = sds_comp[i].comp_info;
        return sds_comp[i].comp_type;
    }
    return 0;
}
/* Allocate and return an array in which to hold an SDS */
char * 
sds_alloc(int32 rank, int32 *dim_sizes, int32 data_type, int32 *npts)
{
    int i, n = 1;
    char *data;
    int element_size;
    for (i = 0; i < rank; i++) {
        n *= dim_sizes[i];
    }
    *npts = n;
    element_size = data_type_size(data_type);
    if ((data = (char *)malloc(n * element_size)) == NULL) {
        fprintf(stderr, "Failed to allocate %d elements of size %d\n", n, data_type);
        exit(21);
    }
    return data;
}
/* Transfer dimension info from one SDS to the other */
int
transfer_dimensions(int32 sds_in, int32 sds_out, int rank, int verbose) {
    int i, dim_in, dim_out;
    char dim_name[1024];
    int32 n_attrs, data_type, dim_size;
    char *scale;

    /* Loop through each dimension up to rank of SDS */
    for (i = 0; i < rank; i++) {
        /* Get dimension handle for input dimension */
        if ((dim_in = SDgetdimid(sds_in, i)) == FAIL) {
            fprintf(stderr, "Failed to get dim_id for dimension %d of sds %d\n", dim_in, sds_in);
            exit(101);
        }
        /* Get dimension handle for output dimension */
        if ((dim_out = SDgetdimid(sds_out, i)) == FAIL) {
            fprintf(stderr, "Failed to get dim_id for dimension %d of sds %d\n", dim_out, sds_out);
            exit(101);
        }
        /* Get dimension information for input dimension */
        if (SDdiminfo(dim_in, dim_name, &dim_size, &data_type, &n_attrs) == FAIL) {
            fprintf(stderr, "Failed to get info for dimension %d of sds %d\n", dim_in, sds_in);
            exit(101);
        }
        /* Set output dimension name */
        SDsetdimname(dim_out, dim_name);
        if (n_attrs && copy_sds_attributes(dim_in, dim_out, n_attrs, verbose) == FAIL) {
            fprintf(stderr, "Failed to copy attributes for dimension %s of sds %d\n", dim_name, sds_in);
            exit(101);
        }
        /* Copy scale information over */
        if (data_type != 0) {
            if (verbose) fprintf(stderr, "Copying scale for dimension %s\n", dim_name);
            if ((scale = (char *)malloc(dim_size * data_type_size(data_type))) == NULL) {
                fprintf(stderr, "Failed to allo %d for dimension scale\n", dim_size);
                exit(102);
            }
            if (SDgetdimscale(dim_in, (void *)scale) == FAIL) {
                fprintf(stderr, "Failed to get scale info for %s\n", dim_name);
                exit(103);
            }
            if (SDsetdimscale(dim_out, dim_size, data_type, (void *)scale) == FAIL) {
                fprintf(stderr, "Failed to set scale info for %s\n", dim_name);
                exit(103);
            }
        }
    }
    return 1;
}
        
int *
transfer_sds(char *infile, char *outfile, sds_comp_t *sds_comp,
    int check_fill, int chunk, int verbose, int32 id_in, int32 id_out)
{
    int32 sd_in, sd_out, sds_in, sds_out;
    int32 dim_sizes[MAX_RANK], edges[MAX_RANK];
    int32 start[MAX_RANK];
    int32 status, rank, data_type, n_attrs;
    int i, j, n_datasets, n_file_attrs, npts;
    int chunkable;
    char *data;
    char sds_name[1024];
    comp_info c_info;
    int32 *refs;
    int compression, compression_info;
    HDF_CHUNK_DEF c_def;
    int chunk_lengths[MAX_RANK];

    /* Open input file */
    sd_in = SDstart(infile, DFACC_READ);
    if (sd_in == FAIL) {
        fprintf(stderr, "Failed to open input %s\n", infile);
        exit(2);
    }
    /* Open output file */
    if ((sd_out = SDstart(outfile, DFACC_WRITE)) == FAIL) {
        fprintf(stderr, "Failed to open output %s\n", outfile);
        exit(3);
    }
    if (SDfileinfo(sd_in, &n_datasets, &n_file_attrs) == FAIL) {
        fprintf(stderr, "Could not get file info from input file\n");
        exit(4);
    }

    /* 
     * Allocate array for pairs of tags for later matching 
     * (one extra for null terminator 
     */
    refs = (int32 *)malloc(2 * (n_datasets + 1) * sizeof(int32));

    copy_sds_attributes(sd_in, sd_out, n_file_attrs, verbose);

    /* Initialize start[] array */
    for (i = 0; i < MAX_RANK; i++) {
        start[i] = 0;
    }

    for (i = 0; i < n_datasets; i++) {
        chunk_lengths[0] = 0;
        chunkable = 0;
        /* Get SDS id */
        if ((sds_in = SDselect(sd_in, i)) == FAIL) {
            fprintf(stderr, "Could not select SDS number %d\n", i);
            exit(5);
        }
        /* Get SDS information */
        if (SDgetinfo(sds_in, sds_name, &rank, dim_sizes, &data_type, &n_attrs) == FAIL) {
            fprintf(stderr, "Could not get dataset info on sds %d\n", sds_in);
            exit(6);
        }
        /* Look up desired compression and chunking info */
        compression = get_compression_info(sds_name, sds_comp, &compression_info, chunk_lengths);
        if (verbose) fprintf(stderr, "Reading in %s...\n", sds_name);

        /* Allocate array to read data into */
        data = sds_alloc(rank, dim_sizes, data_type, &npts);

        /* Set chunking */
        if (chunk_lengths[0]) {
            for (j = 0; j < rank; j++) {
                c_def.chunk_lengths[j] = chunk_lengths[j];
            }
            chunkable = 1;
        }
        else if (chunk && compression) {
            /* Compute reasonable chunk sizes...not! */
            chunkable = compute_chunk_lengths(rank, dim_sizes, c_def.chunk_lengths);
        }
        /* Set edges of SDS */
        for (j = 0; j < rank; j++) {
            edges[j] = dim_sizes[j];
            if (verbose && chunkable) fprintf(stderr, "Chunk in dim %d=%d, size=%d\n", 
                j, c_def.chunk_lengths[j], dim_sizes[j]);
        }

        /* Read data */
        if (SDreaddata (sds_in, start, NULL, edges, (VOIDP)data) == FAIL) {
            fprintf(stderr, "Could not read sds %d\n", sds_in);
            exit(7);
        }
        if (verbose) fprintf(stderr, "Writing out %s...\n", sds_name);

        /* Create output SDS */
        if ((sds_out = SDcreate(sd_out, sds_name, data_type, rank, dim_sizes)) == FAIL) {
            fprintf(stderr, "Failed to create new SDS %s\n", sds_name);
            exit(9);
        }
        /* Match up input and output SDS ids for later vgroup hosting */
        if ((refs[i*2] = SDidtoref(sds_in)) == FAIL) {
            fprintf(stderr, "Cannot get ref for input SDS ID %d\n", sds_in);
            exit(15);
        }
        if ((refs[i*2 + 1] = SDidtoref(sds_out)) == FAIL) {
            fprintf(stderr, "Cannot get ref for input SDS ID %d\n", sds_in);
            exit(15);
        }
        if (verbose) fprintf(stderr, "Input ref = %d, output ref = %d\n", 
            refs[i*2], refs[i*2 + 1]);

        transfer_dimensions(sds_in, sds_out, rank, verbose);

        /* Setup compression factors */
        switch(compression) {
            case 0:
                break;
            case 1:
                c_def.comp.comp_type = compression;
                if (verbose) fprintf(stderr, "Compressing %s with RLE\n", sds_name);
                break;
            case 3:
                c_def.comp.comp_type = compression;
                c_def.comp.cinfo.skphuff.skp_size = compression_info;
                if (verbose) fprintf(stderr, "Compressing %s with Skipping Huffman\n", sds_name);
                break;
            case 4:
                compression_info = (compression_info) ? compression_info : 1;
                c_def.comp.comp_type = compression;
                c_def.comp.cinfo.deflate.level = compression_info;
                if (verbose) fprintf(stderr, "Compressing %s with Gzip level %d\n", 
                    sds_name, compression_info);
                break;
            default:
                fprintf(stderr, "Unrecognized compression code %d\n", compression);
                exit(92);
        }
        /* Setup chunking factors */
        if (chunkable) {
            int compressible = HDF_COMP;
            if (compression == 1 && (data_type == DFNT_FLOAT32 || data_type == DFNT_FLOAT64)) {
                compressible = 0;
            }
            if (verbose) fprintf(stderr, "Compression code %d (%d) with chunks\n",
                    compression, compression_info);
            if (SDsetchunk(sds_out, c_def, (HDF_CHUNK | compressible)) == FAIL) {
                fprintf(stderr, "Failed to set chunks for compression code %d on %s\n",
                    compression, sds_name);
                exit(91);
            }
        }
        /* Set compression parameters */
        else if (compression) {
            if (!(compression == 1 && 
                  (data_type == DFNT_FLOAT32 || data_type == DFNT_FLOAT64))) {
                if (verbose) fprintf(stderr, "Compression code %d (%d)\n",
                    compression, compression_info);
                if (SDsetcompress(sds_out, compression, &c_def.comp.cinfo) == FAIL) {
                    fprintf(stderr, "Failed to set compression code %d on %s\n",
                        compression, sds_name);
                    exit(91);
                }
            }
        }
        /* Finally: write the data */
        if (SDwritedata(sds_out, start, NULL, edges, data) == FAIL) {
            fprintf(stderr, "Failed to write to new SDS %s\n", sds_name);
            exit(10);
        }

        /* Copy over all of the SDS attributes */
        copy_sds_attributes(sds_in, sds_out, n_attrs, verbose);

        /* Try to compute compression ratio */
        if (verbose && compression) {
            int32 before, after;
            double ratio = 0.;
            before = get_sds_size(id_in, sds_in, npts, data_type);
            after = get_sds_size(id_out, sds_out, npts, data_type);
            ratio = (before) ? (double)after/(double)before : -1.;
            fprintf(stderr, "Before=%d  After=%d  Ratio=%.2f\n", 
                before, after, ratio);
        }
        if (SDendaccess(sds_in) == FAIL) {
            fprintf(stderr, "Failed to end access on input SDS %d\n", sds_in);
            exit(8);
        }
        if (SDendaccess(sds_out) == FAIL) {
            fprintf(stderr, "Failed to end access on input SDS %d\n", sds_out);
            exit(11);
        }
        free(data);
    }
    refs[n_datasets * 2] = -1;
    SDend(sd_in);
    SDend(sd_out);
    return refs;
}
int
transfer_vgroups(char *infile, char *outfile, sds_comp_t *sds_comp, 
   int check_fill, int chunk, int verbose) 
{
    int32 in_id, out_id, num_lones;
    int32 *ref_array, *sds_refs;
    int i;
    int32 vg_out;
    if ((in_id = Hopen(infile, DFACC_READ, 0)) == FAIL) {
        fprintf(stderr, "Failed to open input file %s\n", infile);
        exit(11);
    }
    if ((out_id = Hopen(outfile, DFACC_CREATE, 0)) == FAIL) {
        fprintf(stderr, "Failed to open output file %s\n", outfile);
        exit(11);
    }
    if (Vstart (in_id) == FAIL) {
        fprintf(stderr, "Failed to start VGroup interface on %s\n", infile);
        exit(12);
    }
    if (Vstart (out_id) == FAIL) {
        fprintf(stderr, "Failed to start VGroup interface on %s\n", outfile);
        exit(12);
    }
    sds_refs = transfer_sds(infile, outfile, sds_comp, check_fill, chunk, verbose, in_id, out_id);

    num_lones = Vlone(in_id, NULL, 0);
    if (verbose) fprintf(stderr, "%d lone vgroups in %s\n", num_lones, infile);
    ref_array = (int32 *)malloc(num_lones * sizeof(int32));
    num_lones = Vlone(in_id, ref_array, num_lones);
    for (i = 0; i < num_lones; i++) {
        vg_out = transfer_one_vgroup(in_id, out_id, ref_array[i], sds_refs, verbose);
        Vdetach(vg_out);
    }
    transfer_vdatas(in_id, out_id, verbose);
    free(sds_refs);
    Vend(in_id);
    Vend(out_id);
    Hclose(in_id);
    Hclose(out_id);

    return 1;
}
int
transfer_one_vgroup(int32 in_id, int32 out_id, int32 in_ref, 
    int *sds_refs, int verbose) 
{
    char vg_name[1024], vg_class[1024];
    int32 vg_in, vg_out, vg_nest, vdata, out_ref, n_tagrefs, n_pairs;
    int32 *tags, *refs;
    int i;

    vg_in = Vattach (in_id, in_ref, "r");
    if (Vgetclass(vg_in, vg_class) == FAIL) {
        fprintf(stderr, "Failed to get class of vg ref %d\n", in_ref);
        exit(12);
    }
    if (verbose) fprintf(stderr, "Vgroup class = %s\n", vg_class);
    /* Ignore "automatic" ones; those are transferred automatically */
    if (strncmp(vg_class, "CDF0.", 5) == 0) return 1;
    if (strncmp(vg_class, "VAR0.", 5) == 0) return 1;
    if (strncmp(vg_class, "DIM0.", 5) == 0) return 1;

    if (Vgetname(vg_in, vg_name) == FAIL) {
        fprintf(stderr, "Failed to get name of vg ref %d\n", in_ref);
        exit(12);
    }
    if (verbose) fprintf(stderr, "Transferring vgroup name = %s\n", vg_name);
    if ((vg_out = Vattach (out_id, -1, "w")) == FAIL) {
        fprintf(stderr, "Failed to attach to output vg ref\n");
        exit(12);
    }
    if (Vsetclass(vg_out, vg_class) == FAIL) {
        fprintf(stderr, "Failed to set class on %s\n", vg_class);
        exit(12);
    }
    if (Vsetname(vg_out, vg_name) == FAIL) {
        fprintf(stderr, "Failed to set name on %s\n", vg_name);
        exit(12);
    }
    transfer_vgroup_attrs(vg_in, vg_out, verbose);
    n_tagrefs = Vntagrefs(vg_in);
    tags = (int32 *)malloc(n_tagrefs * sizeof(int32));
    refs = (int32 *)malloc(n_tagrefs * sizeof(int32));
    n_pairs = Vgettagrefs(vg_in, tags, refs, n_tagrefs);
    for (i = 0; i < n_pairs; i++) {
        if (Visvg(vg_in, refs[i])) {
            if (verbose) fprintf(stderr, "Nested vgroup %d\n", refs[i]);
            vg_nest = transfer_one_vgroup(in_id, out_id, refs[i], sds_refs, verbose);
            if (Vinsert(vg_out, vg_nest) == FAIL) {
                fprintf(stderr, "Failed to add vg %d to vg %d\n", vg_nest, vg_out);
                exit(12);
            }
            Vdetach(vg_nest);
        }
        else if (Visvs(vg_in, refs[i])) {
            vdata = transfer_one_vdata(in_id, out_id, refs[i], verbose);
            if (Vinsert(vg_out, vdata) == FAIL) {
                fprintf(stderr, "Failed to add vdata %d to vg %d\n", vdata, vg_out);
                exit(12);
            }
            VSdetach(vdata);
        }
        else {
            /* If SDS, find corresponding SDS in output file */
            add_sds_to_vgroup(vg_out, refs[i], sds_refs, verbose);
        }
    }
    return vg_out;
}
int
add_sds_to_vgroup(int32 vgroup, int32 ref, int32 *sds_refs, int verbose) 
{
    int i;
    /* Loop until we get to an sds_ref of -1 */
    for (i = 0; sds_refs[i] >= 0 && sds_refs[i] != ref; i+=2) {
    }

    if (sds_refs[i] < 0) {    /* still not found */
        fprintf(stderr, "SDS reference %d not found in input file\n", ref);
        exit(14);
    }

    if (verbose) fprintf(stderr, "Adding ref %d to vgroup %d\n", sds_refs[i+1], vgroup);
    /* Insert vgroup by sds tag and reference */
    if (Vaddtagref(vgroup, DFTAG_SDS, sds_refs[i+1]) == FAIL) {
        fprintf(stderr, "Failed to add SDS reference %d to vgroup\n", ref);
        exit(14);
    }
    return 1;
}
int
transfer_vdatas(int32 in_id, int32 out_id, int verbose) 
{
    int i, n_lone;
    int32 vdata;
    int32 *refs;
    if ((n_lone = VSlone(in_id, NULL, 0)) == FAIL) {
        fprintf(stderr, "Failed to get num of lone Vdatas\n");
        exit(36);
    }
    if (verbose) fprintf(stderr, "Lone vdatas = %d\n", n_lone);
    refs = (int32 *)malloc(n_lone * sizeof(int32));
    if ((n_lone = VSlone(in_id, refs, n_lone)) == FAIL) {
        fprintf(stderr, "Failed to get refs for lone Vdatas\n");
        exit(36);
    }
    for (i = 0; i < n_lone; i++) {
        vdata = transfer_one_vdata(in_id, out_id, refs[i], verbose);
        VSdetach(vdata);
    }
    return n_lone;
}
int
transfer_one_vdata(int32 in_id, int32 out_id, int32 ref, int verbose) {
    int vdata_in, vdata_out;
    int32 vdata_size, interlace_mode, n_records;
    char fieldname_list[8192];
    char vdata_name[VSNAMELENMAX], vdata_class[VSNAMELENMAX];
    char *field_name;
    int32 field_type, field_order;
    int n_fields, n_attrs;
    int i, j;
    uint8 *buf;

    if ((vdata_in = VSattach (in_id, ref, "r")) == FAIL) {
        fprintf(stderr, "Failed to attach vdata ref %d\n", ref);
        exit(37);
    }
    if (VSgetclass(vdata_in, vdata_class) == FAIL) {
        fprintf(stderr, "Failed to get class for vdata ref %d\n", ref);
        exit(37);
    }
    if (VSinquire(vdata_in, &n_records, &interlace_mode, fieldname_list, 
        &vdata_size, vdata_name) == FAIL) {
        fprintf(stderr, "Failed to get info for vdata ref %d\n", ref);
        exit(37);
    }
    if (verbose) fprintf(stderr, 
        "Transferring vdata %s: class=%s, %d recs, interlace=%d, size=%d\n\tfields='%s'\n",
        vdata_name, vdata_class, n_records, interlace_mode, vdata_size, 
        fieldname_list);
    if ((vdata_out = VSattach (out_id, -1, "w")) == FAIL) {
        fprintf(stderr, "Failed to attach output vdata\n");
        exit(37);
    }
    if (VSsetname(vdata_out, vdata_name) == FAIL) {
        fprintf(stderr, "Failed to set name for output vdata\n");
        exit(37);
    }
    if (VSsetclass(vdata_out, vdata_class) == FAIL) {
        fprintf(stderr, "Failed to set class for output vdata\n");
        exit(37);
    }
    if (VSsetinterlace(vdata_out, interlace_mode) == FAIL) {
        fprintf(stderr, "Failed to set interlace mode for output vdata\n");
        exit(37);
    }
    n_fields = VFnfields(vdata_in);
    if (verbose) fprintf(stderr, "Number of fields = %d\n", n_fields);
    buf = (uint8 *)malloc(n_records * vdata_size);
    for (i = 0; i < n_fields; i++) {
        field_name = VFfieldname(vdata_in, i);
        field_type = VFfieldtype(vdata_in, i);
        field_order = VFfieldorder(vdata_in, i);
        VSfdefine(vdata_out, field_name, field_type, field_order);
        n_attrs = VSfnattrs(vdata_in, i);
        for (j = 0; j < n_attrs; j++) {
            copy_vdata_attribute(vdata_in, vdata_out, i, j, verbose);
        }
    }
    /* Set fields for reading */
    VSsetfields(vdata_in, fieldname_list);
    VSsetfields(vdata_out, fieldname_list);
    n_attrs = VSnattrs(vdata_in);
    for (i = 0; i < n_attrs; i++) {
        copy_vdata_attribute(vdata_in, vdata_out, -1, i, verbose);
    }

    /* Read into buffer, then write */
    if (VSread(vdata_in, buf, n_records, interlace_mode) == FAIL) {
        fprintf(stderr, "Error reading vdata %s\n", vdata_name);
        exit(21);
    }
    if (VSwrite(vdata_out, buf, n_records, interlace_mode) == FAIL) {
        fprintf(stderr, "Error writing vdata %s\n", vdata_name);
        exit(21);
    }

    return vdata_out;
}
int 
transfer_vgroup_attrs(int32 vg_in, int32 vg_out, int verbose) {
    int n_attrs, n_values;
    int32 data_type, size;
    char attr_name[1024];
    int i;
    char *buf;
    /* Get the number of attributes attached to this vgroup.  */
    n_attrs = Vnattrs (vg_in);
    if (verbose) fprintf (stderr, "This vgroup has %d attribute(s)\n", n_attrs);

    for (i = 0; i < n_attrs; i++) {
        Vattrinfo (vg_in, i, attr_name, &data_type, &n_values, &size);
        buf = (char *)malloc(size * n_values);
        Vgetattr (vg_in, i, buf);
        if (verbose) fprintf(stderr, "Copying vg attribute %s\n", attr_name);
        Vsetattr(vg_out, attr_name, data_type, n_values, buf);
        free(buf);
    }
    return 1;
}
int32
get_sds_size(int32 fid, int32 sds_id, int32 npts, int32 data_type) {
    comp_info ci;
    comp_coder_t cc;
    int32 compress_length;
    if (MySDgetcompress(sds_id, &cc, &ci, &compress_length) == FAIL) {
        return (npts * data_type_size(data_type));
    }
    return((cc == COMP_CODE_NONE) ? (npts * data_type_size(data_type)) : compress_length);
}
sds_comp_t *
read_compression_info(char *filename) {
    sds_comp_t *comp;
    char buf[1024];
    FILE *fp;
    int n, i, j, rc, rank;

    if ((fp = fopen(filename, "r")) == (FILE *)NULL) {
        fprintf(stderr, "Cannot open file %s: ", filename);
        perror(NULL);
    }
    fscanf(fp, "%d", &n);
    comp = malloc((n+1) * sizeof(sds_comp_t));
    for (i = 0; i < n; i++) {
        rc = fscanf(fp, "%s %d %d %d", buf, &(comp[i].comp_type), 
            &(comp[i].comp_info), &rank);
        comp[i].sds_name = strdup(buf);
        comp[i].chunk[0] = 0;
        for (j = 0; j < rank; j++) {
            fscanf(fp, "%d", comp[i].chunk+j);
        }
    }
    comp[i].sds_name = NULL;
    comp[i].comp_type = 0;
    comp[i].comp_info = 0;
    fclose(fp);
    return(comp);
}
    
void usage ()
{
    fprintf(stderr, "\
Usage: hdfzip -i input -o output [-h] [-v] [-134 [n] [sds sds ...]] [-f cmp_file] [-C]\n\
  -i input       Input HDF File\n\
  -o output      Output HDF File\n\
  -h             Print usage message\n\
  -v             Verbose: print info to stderr\n\
  -134 n sds     Compression type: 1=RLE, 3=Huffman, 4=gzip\n\
                    n is optional compression info\n\
                    defaults to 1 skip-size for Huffman\n\
                    defaults to deflation level 1 for gzip\n\
                    SDS names following means apply compression only to those SDSs\n\
  -f comp_file   File with compression info in it (instead of above option)\n\
  -C             Apply chunking\n\
");
    exit(1);
}
